/* Client process that communicates with GNU Emacs acting as server.
   Copyright (C) 1986, 1987, 1994, 1999, 2000, 2001
   Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>

#define SCRATCHBOX_CONFIG "/targets/links/scratchbox.config"

/* Like malloc but get fatal error if memory is exhausted.  */

static long *xmalloc(unsigned int size)
{
	long *result;

	result = (long *) malloc (size);
	if (!result) {
		perror("malloc");
		exit(1);
	}

	return result;
}

/* Return a copy of NAME, inserting a &
   before each &, each space, and any initial -.
   Change spaces to underscores, too, so that the
   return value never contains a space.  */

static char *quote_file_name(char *name)
{
	char *copy, *p, *q;

	copy = (char *) xmalloc(strlen(name) * 2 + 1);

	p = name;
	q = copy;

	while (*p) {
		if (*p == ' ') {
			*q++ = '&';
			*q++ = '_';
			p++;
		} else {
			if (*p == '&' || (*p == '-' && p == name)) {
				*q++ = '&';
			}
			*q++ = *p++;
		}
	}

	*q++ = 0;

	return copy;
}

/* Three possibilities:
   2 - can't be `stat'ed		(sets errno)
   1 - isn't owned by us
   0 - success: none of the above */

static int socket_status(char *socket_name)
{
	struct stat statbfr;

	if (stat(socket_name, &statbfr) == -1) {
		return 2;
	}

	if (statbfr.st_uid != geteuid()) {
		return 1;
	}

	return 0;
}

static char *get_sbox_root(void)
{
	int fd;
	char path[PATH_MAX];

	fd = open(SCRATCHBOX_CONFIG, O_RDONLY);
	if (fd < 0) {
		return NULL;
	}
	close(fd);

	strcpy(path, "/scratchbox/users/");
	strcat(path, getenv("USER"));

	return quote_file_name(path);
}

int main(int argc, char **argv)
{
	char *system_name, *cwd, *root, string[BUFSIZ];
	int system_name_length, s, i, sock_status;
	FILE *out;
	struct sockaddr_un server;

	if (argc < 2) {
		return 0;
	}

	/* Open up an AF_UNIX socket in /tmp directory */

	if ((s = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}

	server.sun_family = AF_UNIX;
	system_name_length = 32;

	while (1) {
		system_name = (char *) xmalloc(system_name_length + 1);

		/* system_name must be null-terminated string.  */
		system_name[system_name_length] = '\0';

		if (gethostname(system_name, system_name_length) == 0) {
			break;
		}

		free(system_name);
		system_name_length *= 2;
	}

	sprintf(server.sun_path, "/tmp/esrv%d-%s",
		(int) geteuid(), system_name);

	/* See if the socket exists, and if it's owned by us. */

	sock_status = socket_status(server.sun_path);
	if (sock_status) {
		/* Failing that, see if LOGNAME or USER exist and differ from
		   our euid.  If so, look for a socket based on the UID
		   associated with the name.  This is reminiscent of the logic
		   that init_editfns uses to set the global Vuser_full_name. */

		char *user_name = (char *) getenv("LOGNAME");
		if (!user_name) {
			user_name = (char *) getenv("USER");
		}

		if (user_name) {
			struct passwd *pw = getpwnam(user_name);
			if (pw && (pw->pw_uid != geteuid())) {
				/* We're running under su, apparently. */
				sprintf(server.sun_path, "/tmp/esrv%d-%s",
					(int) pw->pw_uid, system_name);
				sock_status = socket_status(server.sun_path);
			}
		}
	}

	switch (sock_status) {
	case 1:
		/* There's a socket, but it isn't owned by us.  This is OK if
		   we are root. */

		if (0 != geteuid()) {
			fprintf(stderr, "Invalid socket owner\n");
			exit (1);
		}
		break;

	case 2:
		fprintf(stderr, "Can't stat %s\n", server.sun_path);
		exit(1);
		break;
	}

	if (connect(s, (struct sockaddr *) &server,
		    strlen(server.sun_path) + 2) < 0) {
		perror("connect");
		exit(1);
	}

	/* We use the stream OUT to send our command to the server.  */

	if (!(out = fdopen(s, "r+"))) {
		perror("fdopen");
		exit(1);
	}

	cwd = getcwd(string, sizeof (string));
	if (!cwd) {
		perror("getcwd");
		exit(1);
	}

	fprintf(out, "-nowait ");

	root = get_sbox_root();

	for (i = 1; i < argc; i++) {
		char buf[PATH_MAX], *dir;

		if (root) {
			fprintf(out, "%s", root);
		}

		dir = realpath(dirname(strdup(argv[i])), buf);
		if (!dir) {
			perror("realpath");
			exit(1);
		}

		fprintf(out, "%s/%s ", quote_file_name(dir),
			quote_file_name(basename(argv[i])));
	}
	fprintf(out, "\n");
	fflush(out);

	return 0;
}
