/*
 * Copyright (C) 2003-2004 Nokia
 * Copyright (C) 2001      BitWagon Software LLC
 * Copyright (C) 1993-2003 Free Software Foundation
 *
 * This file may be used, modified, and distributed under the terms
 * of the GNU General Public License version 2.
 */

#define _GNU_SOURCE

#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/syscall.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <libgen.h>
#include <limits.h>
#include <errno.h>
#include <paths.h>

#include <sys/utsname.h>
#include <dirent.h>
#include <sys/stat.h>

#include <elf.h>
#include <sys/mman.h>

#include <sys/time.h>
#include <time.h>

/*
 * execve wrapping is taken directly from fakeroot
 * please look at fakeroot source code for examples, documentation
 * and overall understanding how this works :-)
 * */


#define RPATH_PREFIX "/scratchbox/"

#define LIBCPATH "/scratchbox/host_shared/lib/libc.so.6"

#define BUFFERSIZE PATH_MAX

#ifdef DEBUG
#define DEBUG_OUT(...) fprintf(stderr, __VA_ARGS__)
#else
#define DEBUG_OUT(...)
#endif


extern char **environ;

struct next_wrap_st {
	void **doit;
	char *name;
};

static void *get_libc(void)
{
#ifndef RTLD_NEXT
	static void *lib = 0;
	if (!lib) {
		lib = dlopen(LIBCPATH, RTLD_LAZY);
	}
	if (NULL == lib) {
		fprintf(stderr, "Couldn't find libc at: %s\n", LIBCPATH);
		abort();
	}
	return lib;
#else
	return RTLD_NEXT;
#endif
}

static void load_library_symbols(void);

int (*next_execve)(const char *, char *const [], char *const []);
int (*next_uname)(struct utsname *buf);

static int tmp_execve(const char *file, char *const argv[], char *const envp[])
{
	load_library_symbols();
	return next_execve(file, argv, envp);
}

static int tmp_uname(struct utsname *buf)
{
	load_library_symbols();
	return next_uname(buf);
}

int (*next_execve)(const char *, char *const [], char *const []) = tmp_execve;
int (*next_uname)(struct utsname *) = tmp_uname;

static struct next_wrap_st next_wrap[] = {
	{ (void (*)) &next_execve, "execve" },
	{ (void (*)) &next_uname,  "uname"  },
	{ NULL, NULL }
};

static void load_library_symbols(void)
{
	static int done = 0;
	int i;

	if (done) {
		return;
	}

	for (i = 0; next_wrap[i].doit; i++) {
		char *msg;

		*(next_wrap[i].doit) = dlsym(get_libc(), next_wrap[i].name);
		if ((msg = dlerror())) {
			fprintf(stderr, "dlsym(%s): %s\n",
				next_wrap[i].name, msg);
		}
	}
}


enum { LOG_NONE, LOG_DIRECTORY, LOG_BINARY, LOG_IGNORE };

static void print_log(const char *from,
		      const char *to,
		      int action,
		      int hash_bang)
{
	char *filename, *hbstr;
	FILE *file;
	struct timeval tv;
	struct tm *t;

	hbstr = "  ";
	if (hash_bang) {
		hbstr = "#!";
	}

	gettimeofday(&tv, NULL);
	t = gmtime(&tv.tv_sec);

	filename = getenv("SBOX_REDIRECT_LOG");
	if (!filename || filename[0] == '\0') {
		return;
	}

	file = fopen(filename, "a");
	if (!file) {
		fprintf(stderr, "libsb: %s: %s\n", filename, strerror(errno));
		return;
	}

	fprintf(file, "[%02d.%02d.%04d %02d:%02d:%02d %5d] ",
		t->tm_mday, t->tm_mon + 1, 1900 + t->tm_year,
		t->tm_hour, t->tm_min, t->tm_sec, getpid());

	switch (action) {
	case LOG_NONE:
		fprintf(file, "NOT %s \"%s\"\n", hbstr, from);
		break;

	case LOG_DIRECTORY:
		fprintf(file, "DIR %s \"%s\" -> \"%s\"\n", hbstr, from, to);
		break;

	case LOG_BINARY:
		fprintf(file, "BIN %s \"%s\" -> \"%s\"\n", hbstr, from, to);
		break;

	case LOG_IGNORE:
		fprintf(file, "IGN %s \"%s\"\n", hbstr, from);
		break;
	}

	fclose(file);
}


#define PAGE_MASK (~0u << 12)

#if __BYTE_ORDER == __BIG_ENDIAN
# define elf_endianness ELFDATA2MSB
#elif __BYTE_ORDER == __LITTLE_ENDIAN
# define elf_endianness ELFDATA2LSB
#else
# error Invalid __BYTE_ORDER
#endif

/* TODO: should we drop class testing? */
#define elf_class   ELFCLASS32

/* TODO: how can these be made portable? */
#define elf_machine EM_386

/*
 * Returns true if the file is dynamic ELF for the native architecture
 * without -rpath pointing to /scratchbox/host_shared/lib.
 */
static int should_drop_preload(const char *filename)
{
	int drop, fd, phnum, j;
	struct stat status;
	char *region;
	unsigned int reloc0, ph_base, ph_frag;
	Elf32_Ehdr *ehdr;
	Elf32_Phdr *phdr;

	drop = 0;

	fd = open(filename, O_RDONLY, 0);
	if (fd < 0) {
		goto _out;
	}

	if (fstat(fd, &status) < 0) {
		goto _out_close;
	}

	if (!S_ISREG(status.st_mode) && !S_ISLNK(status.st_mode)) {
		goto _out_close;
	}

	region = mmap(0, status.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (!region) {
		goto _out_close;
	}

	ehdr = (Elf32_Ehdr *) region;

	if (strncmp((char *) ehdr, ELFMAG, SELFMAG) != 0) {
		goto _out_munmap;
	}

	drop = 1;

	if (ehdr->e_ident[EI_DATA] != elf_endianness ||
	    ehdr->e_ident[EI_CLASS] != elf_class ||
	    ehdr->e_ident[EI_VERSION] != EV_CURRENT ||
	    ehdr->e_machine != elf_machine) {
		goto _out_munmap;
	}

	phnum = ehdr->e_phnum;
	reloc0 = ~0;
	ph_base = ehdr->e_phoff & PAGE_MASK;
	ph_frag = ehdr->e_phoff - ph_base;

	phdr = (Elf32_Phdr *) (region + ph_base + ph_frag);

	for (j = phnum; --j >= 0; ++phdr) {
		if (PT_LOAD == phdr->p_type && ~0 == reloc0) {
			reloc0 = phdr->p_vaddr - phdr->p_offset;
		}
	}

	phdr -= phnum;

	for (j = phnum; --j >= 0; ++phdr) {
		int szDyn;
		unsigned int pt_base, pt_frag;
		Elf32_Dyn *dp_rpath, *dp_strsz, *dp_strtab, *dp;

		if (PT_DYNAMIC != phdr->p_type) {
			continue;
		}

		pt_base = phdr->p_offset & PAGE_MASK;
		pt_frag = phdr->p_offset - pt_base;
		szDyn = phdr->p_filesz;

		dp_rpath = NULL;
		dp_strsz = NULL;
		dp_strtab = NULL;

		dp = (Elf32_Dyn *) ((char *) region + pt_base + pt_frag);

		for (; 0 <= (szDyn -= sizeof (*dp)); ++dp) {
			if (DT_RPATH == dp->d_tag) {
				dp_rpath = dp;
			} else if (DT_STRTAB == dp->d_tag) {
				dp_strtab = dp;
			} else if (DT_STRSZ == dp->d_tag) {
				dp_strsz = dp;
			}
		}

		if (dp_rpath && dp_strtab && dp_strsz) {
			char *rpathstr, *strtab;
			unsigned int unloc, str_base, str_frag;

			unloc = dp_strtab->d_un.d_val - reloc0;
			str_base = unloc & PAGE_MASK;
			str_frag = unloc - str_base;

			strtab = (char *) (region + str_base + str_frag);
			rpathstr = dp_rpath->d_un.d_val + strtab;

			if (strncmp(rpathstr, RPATH_PREFIX,
				    strlen(RPATH_PREFIX)) == 0) {
				drop = 0;
			}
		}
	}

_out_munmap:
	munmap(region, status.st_size);
_out_close:
	close(fd);
_out:
	return drop;
}

static char *const *drop_preload(char *const oldenv[])
{
	char **newenv;
	int i, j;

	for (i = 0; oldenv[i]; ++i);

	newenv = (char **) calloc(i + 1, sizeof (char *));
	if (!newenv) {
		fprintf(stderr, "libsb: %s\n", strerror(errno));
		return oldenv;
	}

	for (i = 0, j = 0; oldenv[i]; ++i) {
		if (strncmp(oldenv[i], "LD_PRELOAD=", 11) != 0) {
			newenv[j++] = oldenv[i];
		}
	}

	return newenv;
}

/*
 * Checks if a file has #! and reads the interpreter and its argument to
 * buffers of 1024 bytes.
 */
static int check_for_hashbang(const char *filename,
			      char *interp_buf,
			      char *interparg_buf)
{
	int hash_bang = 0;
	FILE *fp;

	if ((fp = fopen(filename, "r"))) {
		char buf[BUFFERSIZE];  /* hash bang detection buffer */

		memset(buf, 0, BUFFERSIZE);
		fread(buf, BUFFERSIZE - 1, 1, fp);

		if (buf[0] == '#' && buf[1] == '!') {
			int begin, end;

			hash_bang = 1;

			/* find begin of interpreter */
			begin = 2;
			while (buf[begin] == ' ' || buf[begin] == '\t') {
				begin++;
			}

			/* find end of interpreter */
			end = begin;
			while (buf[end] != ' '  && buf[end] != '\t' &&
			       buf[end] != '\0' && buf[end] != '\n') {
				end++;
			}

			/* get the interpreter */
			strncpy(interp_buf, &buf[begin], end - begin);

			/* find begin of argument */
			while (buf[end] == ' ' || buf[end] == '\t') {
				end++;
			}
			begin = end;

			/* find end of argument */
			while (buf[end] != '\n' && buf[end] != '\0') {
				end++;
			}

			/* get the interpreter argument */
			strncpy(interparg_buf, &buf[begin], end - begin);
		}

		fclose(fp);
	}

	return hash_bang;
}

static void prepare_hashbang_call(const char *file,
				  char *const argv[],
				  char *interpreter,
				  char *interpreter_arg,
				  char ***real_argv)
{
	char **tmp, **tmp2;
	int c;

	tmp = (char **) argv;
	c = 0;
	while (*tmp) {
		c++;
		tmp++;
	}
	*real_argv = malloc(sizeof (char **) * (c + 3));
	tmp = (char **) argv;
	tmp2 = (char **) *real_argv;
	*tmp2++ = (char *) file;
	if (interpreter_arg[0] != '\0') {
		*tmp2++ = interpreter_arg;
	}

	*tmp2++ = (char *) file;

	tmp++;  // skip argv[0]
	while (*tmp) {
		*tmp2++ = *tmp++;
	}
	*tmp2 = NULL;
}

/*
 * Checks if the directory part of bin_path is under an entry in bindir_list.
 */
static int in_bindirs(const char *bindir_list,
		      const char *bin_path)
{
	char *list, *dir, *bindir;

	list = strdup(bindir_list);
	dir = dirname(strdup(bin_path));

	while ((bindir = strsep(&list, ":"))) {
		if (strcmp(bindir, dir) == 0) {
			return 1;
		}
	}

	return 0;
}

/*
 * Concatenates PATH and FILE into BUF and returns 0. If the length of the
 * result string is over PATH_MAX, -1 will be returned and message is printed
 * to stderr.
 */
static int build_path(const char *path,
		      const char *file,
		      char *buf)
{
	if (strlen(path) + strlen(file) + 2 > PATH_MAX) {
		fprintf(stderr, "libsb/execve: Path is too long: %s/%s",
			path, file);
		return -1;
	}

	strcpy(buf, path);
	strcat(buf, "/");
	strcat(buf, file);
	return 0;
}

static int do_exec(const char *file,
		   char *const argv[],
		   char *const envp_orig[])
{
	char *const *envp = envp_orig;

	if (should_drop_preload(file)) {
		envp = drop_preload(envp_orig);
	}

#ifdef MEMCHECK
	exit(0);
#endif

	return next_execve(file, argv, envp);
}

int execve(const char *file,
	   char *const argv[],
	   char *const envp[])
{
	char *bin, *redir_list, *ignore_list;
	char *fromdir_list, *todir_list, *bindir;
	char interpreter[BUFFERSIZE];
	char interpreter_arg[BUFFERSIZE];
	int hash_bang = 0, log_action = LOG_NONE;

	DEBUG_OUT("fake execve!\n");

	bin = basename(strdup(file));

	fromdir_list = getenv("SBOX_REDIRECT_FROM_DIRS");
	todir_list = getenv("SBOX_REDIRECT_TO_DIRS");
	redir_list = getenv("SBOX_REDIRECT_BINARIES");
	ignore_list = getenv("SBOX_REDIRECT_IGNORE");

	if (!redir_list && (!fromdir_list || !todir_list)) {
		/* nothing to redirect */
		goto default_handler;
	}

	/* ignore individual binaries */

	if (ignore_list) {
		char *path;

		ignore_list = strdup(ignore_list);

		while ((path = strsep(&ignore_list, ":"))) {
			if (strcmp(file, path) == 0) {
				log_action = LOG_IGNORE;
				goto default_handler;
			}
		}
	}

	memset(interpreter, 0, BUFFERSIZE);
	memset(interpreter_arg, 0, BUFFERSIZE);
	hash_bang = check_for_hashbang(file, interpreter, interpreter_arg);

	/* redirect individual binaries */

	if (redir_list) {
		char *li, **my_argv, *tmp_file;

		redir_list = strdup(redir_list);

		if (hash_bang) {
			tmp_file = interpreter;
			prepare_hashbang_call(file, argv, interpreter, interpreter_arg, &my_argv);
		} else {
			tmp_file = (char *) file;
			my_argv = (char **) argv;
		}

		while ((li = strsep(&redir_list, ","))) {
			char *exe1, *exe2;

			exe1 = strsep(&li, ":");
			if (!exe1) {
				continue;
			}

			exe2 = strsep(&li, ":");
			if (!exe2) {
				continue;
			}

			if (strcmp(exe1, tmp_file) == 0) {
				my_argv[0] = exe2;

				print_log(exe1, exe2, LOG_BINARY, hash_bang);

				return do_exec(exe2, my_argv, envp);
			}
		}
	}

	/* redirect everything under directories listed in fromdir_list */

	if (!fromdir_list || !todir_list) {
		goto default_handler;
	}
	if (!in_bindirs(fromdir_list, file)) {
		if (!hash_bang) {
			goto default_handler;
		}
		if (!in_bindirs(fromdir_list, interpreter)) {
			goto default_handler;
		}
	}

	todir_list = strdup(todir_list);

	while ((bindir = strsep(&todir_list, ":"))) {
		char fullname[PATH_MAX];
		char **my_argv;
		struct stat f;

		DEBUG_OUT("trying to redirect to %s\n", bindir);

		if (hash_bang) {
			char *inter_base = basename(strdup(interpreter));

			if (build_path(bindir, inter_base, fullname) < 0) {
				continue;
			}

			if (stat(fullname, &f) != 0) {
				continue;
			}

			prepare_hashbang_call(file, argv, fullname, interpreter_arg, &my_argv);
			my_argv[0] = fullname;

			print_log(interpreter, fullname, LOG_DIRECTORY, hash_bang);
		} else {
			if (build_path(bindir, bin, fullname) < 0) {
				continue;
			}

			if (stat(fullname, &f) != 0) {
				DEBUG_OUT(strerror(errno));
				continue;
			}

			DEBUG_OUT("stat OK\n");

			if (!S_ISREG(f.st_mode)) {
				DEBUG_OUT("Not a regular file");
				continue;
			}

			my_argv = (char **) argv;

			print_log(file, fullname, LOG_DIRECTORY, hash_bang);
		}

		return do_exec(fullname, my_argv, envp);
	}

	/* fall back to default behaviour */

default_handler:
	DEBUG_OUT("default handler\n");
	print_log(file, NULL, log_action, hash_bang);

	return do_exec(file, argv, envp);
}


static void set_name(char *str, const char *name)
{
	char *value = getenv(name);
	if (value) {
		strncpy(str, value, _UTSNAME_LENGTH);
	}
}

int uname(struct utsname *buf)
{
	if (next_uname(buf) < 0) {
		return -1;
	}

	set_name(buf->sysname,  "SBOX_UNAME_SYSNAME");
	set_name(buf->nodename, "SBOX_UNAME_NODENAME");
	set_name(buf->release,  "SBOX_UNAME_RELEASE");
	set_name(buf->version,  "SBOX_UNAME_VERSION");
	set_name(buf->machine,  "SBOX_UNAME_MACHINE");

	return 0;
}






/* ************ exec family of functions copied from glibc-2.3.2 ***********/


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with all arguments after PATH until
   a NULL pointer and environment from `environ'.  */
int
execl (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execve (path, (char *const *) argv, __environ);
}




/* Execute PATH with all arguments after PATH until a NULL pointer,
   and the argument after that for environment.  */
int
execle (const char *path, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  const char *const *envp;
  unsigned int i;
  va_list args;
  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }

  envp = va_arg (args, const char *const *);
  va_end (args);

  return execve (path, (char *const *) argv, (char *const *) envp);
}




/* Execute FILE, searching in the `PATH' environment variable if
   it contains no slashes, with all arguments after FILE until a
   NULL pointer and environment from `environ'.  */
int
execlp (const char *file, const char *arg, ...)
{
  size_t argv_max = 1024;
  const char **argv = alloca (argv_max * sizeof (const char *));
  unsigned int i;
  va_list args;

  argv[0] = arg;

  va_start (args, arg);
  i = 0;
  while (argv[i++] != NULL)
    {
      if (i == argv_max)
	{
	  const char **nptr = alloca ((argv_max *= 2) * sizeof (const char *));

#ifndef _STACK_GROWS_UP
	  if ((char *) nptr + argv_max == (char *) argv)
	    {
	      /* Stack grows down.  */
	      argv = (const char **) memcpy (nptr, argv,
					     i * sizeof (const char *));
	      argv_max += i;
	    }
	  else
#endif
#ifndef _STACK_GROWS_DOWN
	    if ((char *) argv + i == (char *) nptr)
	    /* Stack grows up.  */
	    argv_max += i;
	  else
#endif
	    /* We have a hole in the stack.  */
	    argv = (const char **) memcpy (nptr, argv,
					   i * sizeof (const char *));
	}

      argv[i] = va_arg (args, const char *);
    }
  va_end (args);

  return execvp (file, (char *const *) argv);
}


#ifndef	HAVE_GNU_LD
# define __environ	environ
#endif

/* Execute PATH with arguments ARGV and environment from `environ'.  */
int
execv (const char *path, char *const argv[])
{
  return execve (path, argv, __environ);
}




/* The file is accessible but it is not an executable file.  Invoke
   the shell to interpret it as a script.  */
static void
sb_script_execute (const char *file, char *const argv[])
{
  /* Count the arguments.  */
  int argc = 0;
  while (argv[argc++])
    ;

  /* Construct an argument list for the shell.  */
  {
    char *new_argv[argc + 1];
    new_argv[0] = (char *) _PATH_BSHELL;
    new_argv[1] = (char *) file;
    while (argc > 1)
      {
	new_argv[argc] = argv[argc - 1];
	--argc;
      }

    /* Execute the shell.  */
    execve (new_argv[0], new_argv, __environ);
  }
}


/* Execute FILE, searching in the `PATH' environment variable if it contains
   no slashes, with arguments ARGV and environment from `environ'.  */
int
execvp (file, argv)
     const char *file;
     char *const argv[];
{
  if (*file == '\0')
    {
      /* We check the simple case first. */
      errno = ENOENT;
      return -1;
    }

  if (strchr (file, '/') != NULL)
    {
      /* Don't search when it contains a slash.  */
      execve (file, argv, __environ);

      if (errno == ENOEXEC)
	sb_script_execute (file, argv);
    }
  else
    {
      int got_eacces = 0;
      char *path, *p, *name;
      size_t len;
      size_t pathlen;

      path = getenv ("PATH");
      if (path == NULL)
	{
	  /* There is no `PATH' in the environment.
	     The default search path is the current directory
	     followed by the path `confstr' returns for `_CS_PATH'.  */
	  len = confstr (_CS_PATH, (char *) NULL, 0);
	  path = (char *) alloca (1 + len);
	  path[0] = ':';
	  (void) confstr (_CS_PATH, path + 1, len);
	}

      len = strlen (file) + 1;
      pathlen = strlen (path);
      name = alloca (pathlen + len + 1);
      /* Copy the file name at the top.  */
      name = (char *) memcpy (name + pathlen + 1, file, len);
      /* And add the slash.  */
      *--name = '/';

      p = path;
      do
	{
	  char *startp;

	  path = p;
	  p = strchrnul (path, ':');

	  if (p == path)
	    /* Two adjacent colons, or a colon at the beginning or the end
	       of `PATH' means to search the current directory.  */
	    startp = name + 1;
	  else
	    startp = (char *) memcpy (name - (p - path), path, p - path);

	  /* Try to execute this name.  If it works, execv will not return.  */
	  execve (startp, argv, __environ);

	  if (errno == ENOEXEC)
	    sb_script_execute (startp, argv);

	  switch (errno)
	    {
	    case EACCES:
	      /* Record the we got a `Permission denied' error.  If we end
		 up finding no executable we can use, we want to diagnose
		 that we did find one but were denied access.  */
	      got_eacces = 1;
	    case ENOENT:
	    case ESTALE:
	    case ENOTDIR:
	      /* Those errors indicate the file is missing or not executable
		 by us, in which case we want to just try the next path
		 directory.  */
	      break;

	    default:
	      /* Some other error means we found an executable file, but
		 something went wrong executing it; return the error to our
		 caller.  */
	      return -1;
	    }
	}
      while (*p++ != '\0');

      /* We tried every element and none of them worked.  */
      if (got_eacces)
	/* At least one failure was due to permissions, so report that
           error.  */
	errno = EACCES;
    }

  /* Return the error from the last attempt (probably ENOENT).  */
  return -1;
}


#include	<sys/types.h>
#include	<sys/wait.h>
#include	<errno.h>
#include	<unistd.h>

int
system(const char *cmdstring)	/* version without signal handling */
{
	pid_t	pid;
	int		status;

	if (cmdstring == NULL)
		return(1);		/* always a command processor with Unix */

	if ( (pid = fork()) < 0) {
		status = -1;	/* probably out of processes */

	} else if (pid == 0) {				/* child */
		execl("/bin/sh", "sh", "-c", cmdstring, (char *) 0);
		_exit(127);		/* execl error */

	} else {							/* parent */
		while (waitpid(pid, &status, 0) < 0)
			if (errno != EINTR) {
				status = -1; /* error other than EINTR from waitpid() */
				break;
			}
	}

	return(status);
}

#include	<sys/wait.h>
#include	<errno.h>
#include	<fcntl.h>

static pid_t	*childpid = NULL;
						/* ptr to array allocated at run-time */
static int		maxfd;	/* from our open_max(), {Prog openmax} */

#define	SHELL	"/scratchbox/tools/bin/bash"

FILE *
popen(const char *cmdstring, const char *type)
{
	int		i, pfd[2];
	pid_t	pid;
	FILE	*fp;

			/* only allow "r" or "w" */
	if ((type[0] != 'r' && type[0] != 'w') || type[1] != 0) {
		errno = EINVAL;		/* required by POSIX.2 */
		return(NULL);
	}

	if (childpid == NULL) {		/* first time through */
				/* allocate zeroed out array for child pids */
		maxfd = FOPEN_MAX;
		if ( (childpid = calloc(maxfd, sizeof(pid_t))) == NULL)
			return(NULL);
	}

	if (pipe(pfd) < 0)
		return(NULL);	/* errno set by pipe() */

	if ( (pid = fork()) < 0)
		return(NULL);	/* errno set by fork() */
	else if (pid == 0) {							/* child */
		if (*type == 'r') {
			close(pfd[0]);
			if (pfd[1] != STDOUT_FILENO) {
				dup2(pfd[1], STDOUT_FILENO);
				close(pfd[1]);
			}
		} else {
			close(pfd[1]);
			if (pfd[0] != STDIN_FILENO) {
				dup2(pfd[0], STDIN_FILENO);
				close(pfd[0]);
			}
		}
			/* close all descriptors in childpid[] */
		for (i = 0; i < maxfd; i++)
			if (childpid[i] > 0)
				close(i);

		execl(SHELL, "sh", "-c", cmdstring, (char *) 0);
		_exit(127);
	}
								/* parent */
	if (*type == 'r') {
		close(pfd[1]);
		if ( (fp = fdopen(pfd[0], type)) == NULL)
			return(NULL);
	} else {
		close(pfd[0]);
		if ( (fp = fdopen(pfd[1], type)) == NULL)
			return(NULL);
	}
	childpid[fileno(fp)] = pid;	/* remember child pid for this fd */
	return(fp);
}

int
pclose(FILE *fp)
{
	int		fd, stat;
	pid_t	pid;

	if (childpid == NULL)
		return(-1);		/* popen() has never been called */

	fd = fileno(fp);
	if ( (pid = childpid[fd]) == 0)
		return(-1);		/* fp wasn't opened by popen() */

	childpid[fd] = 0;
	if (fclose(fp) == EOF)
		return(-1);

	while (waitpid(pid, &stat, 0) < 0)
		if (errno != EINTR)
			return(-1);	/* error other than EINTR from waitpid() */

	return(stat);	/* return child's termination status */
}
